from Feature_Calculation_stripped import *
from math import *
from sklearn.ensemble import RandomForestClassifier
from scipy import ndimage
from scipy import spatial
import copy
from sklearn.externals import joblib


""" applies tracking algorithm """


#
#   FUNCTIONS
#


def classes2overlays(pixelclassses, height, width):
    color_lookup_yes=np.array([black,yes_color])
    color_lookup_no=np.array([no_color,black])
    pixel_colors_yes=color_lookup_yes[pixelclassses.astype(int)]
    pixel_colors_no=color_lookup_no[pixelclassses.astype(int)]
    yes_overlay=np.reshape(pixel_colors_yes,(height, width, 3)).astype('uint8')
    no_overlay=np.reshape(pixel_colors_no,(height, width, 3)).astype('uint8')
    return (yes_overlay, no_overlay)


def superimpose(image, overlay):
    img2gray = cv2.cvtColor(overlay,cv2.COLOR_BGR2GRAY)
    ret, mask = cv2.threshold(img2gray, 10, 255, cv2.THRESH_BINARY)
    mask_inv = cv2.bitwise_not(mask)
    img1_bg = cv2.bitwise_and(image,image,mask = mask_inv)
    img2_fg = cv2.bitwise_and(overlay,overlay,mask = mask)
    result = cv2.add(img1_bg,img2_fg)
    return result


def classificationImage(image, pixelclasses, foreground=True, background=False):
    (h,w)=image.shape[:2]
    (yes_overlay, no_overlay) = classes2overlays(pixelclasses, h, w)
    if foreground:
        image = superimpose(image, yes_overlay)
    if background:
        image = superimpose(image, no_overlay)
    return image


def visual_classifier_test_vectorized(filename, predictor, foreground=True, background=False):
    image = cv2.imread(filename+' original.png')
    if image is None:
        raise FileNotFoundError(imagefile_name)
    compute_feature_image(filename)
    feature_image=np.load(filename+' features.npy',None)
    (h,w)=image.shape[:2]
    feature_pixel_list = np.reshape(feature_image,(h*w,num_features))
    pixelclasses=predictor.predict(feature_pixel_list)
    image = classificationImage(image, pixelclasses, foreground, background)
    return image


def bresenham_line((xs,ys),(xe,ye)):
    x,y = int(round(xs)), int(round(ys))
    dx, dy = (xe-xs), (ye-ys)
    if abs(dy)>abs(dx):
        return [(x,y) for (y,x) in bresenham_line((ys,xs),(ye,xe))]
    elif xe<xs:
        return bresenham_line((xe,ye),(xs,ys))
    else:
        error = -dx
        pts = []
        for x in range(xs, xe+1):
            pts.append((x,y))
            error += 2*abs(dy)
            if error >= 0.0:
                y += int(copysign(1,ye-ys))
                error -= 2*dx
        return pts


def path_score(image_gray, start, end, function=np.mean, debug=False):
    scores = [image_gray[x,y] for (y, x) in bresenham_line(start,end)]
    overall_score = function(scores)
    return overall_score


def to_int_tuple((x,y)):
    return (int(round(x)),int(round(y)))


def rejoin_keypoints(image, ant_coords, debug=False):
    max_distance = JOIN_DISTANCE
    threshold = 160
    keypoints = copy.copy(ant_coords)
    ants_kd = spatial.KDTree(keypoints,5)
    close_keypoint_pairs = ants_kd.query_pairs(max_distance)
    to_be_joined = []
    for (a,b) in close_keypoint_pairs:
        posA = tuple(ants_kd.data[a])
        posB = tuple(ants_kd.data[b])
        if path_score(image, to_int_tuple(posA), to_int_tuple(posB), function=np.mean, debug=debug) > threshold :
            to_be_joined.append((posA, posB))
    image = cv2.bitwise_not(image)
    for (posA,posB) in to_be_joined:
        (x1, y1) = posA
        (x2, y2) = posB
        cv2.line(image, to_int_tuple(posA), to_int_tuple(posB), (0, 0, 255))
        if posA in keypoints: keypoints.remove(posA)
        cv2.circle(image, to_int_tuple(posA), 4, (0, 255, 0), 0)
        if posB in keypoints: keypoints.remove(posB)
        cv2.circle(image, to_int_tuple(posB), 4, (0, 255, 0), 0)
        keypoints.append(((x1 + x2) / 2, (y1 + y2) /2))
    return keypoints


def find_ant_coordinates_diy(filename, predictor, background_predictor=None, debug=False):

    def probs2gray(x):
        low = 0
        high = 255
        [a, b] = x
        return int(round(a * low + b * high))

    image = cv2.imread(filename+' original.png')
    if image is None:
        raise FileNotFoundError(filename+' original.png')
    compute_feature_image(filename)
    feature_image=np.load(filename+' features.npy',None)
    if feature_image is None:
        raise FileNotFoundError(filename+' features.npy')
    (h, w) = image.shape[:2]

    num_features = len(feature_vector((1,1)))
    feature_pixel_list = np.reshape(feature_image,(h*w, num_features))
    pixelprobs=predictor.predict_proba(feature_pixel_list)
    x=np.apply_along_axis(probs2gray,1,pixelprobs)
    img=np.reshape(x.astype('uint8'),(h,w))

    if background_predictor:
        compute_feature_image(filename, narrow=True)
        feature_narrow_image=np.load(filename+' narrow features.npy',None)
        if feature_narrow_image is None:
            raise FileNotFoundError(filename+' narrow features.npy')
        num_features=len(feature_vector((1,1), narrow=True))
        feature_narrow_pixel_list = np.reshape(feature_narrow_image,(h*w, num_features))
        pixel_is_background=background_predictor.predict(feature_narrow_pixel_list)
        pixel_is_not_background=np.absolute(np.add(pixel_is_background,-1))
        bg_img=np.reshape(pixel_is_not_background.astype('uint8'),(h,w))

        if debug:
            cvShow(np.multiply(255,bg_img),title="background")

            fg_img=np.reshape(pixel_is_background.astype('uint8'),(h,w))
            modified=np.multiply(img,fg_img)
            cvShow(modified, title="eliminated background probs")

        img=np.multiply(img,bg_img)


    cv2.imwrite(filename+' probabilities.png',img)

    if debug:
        cvShow(img,title='probabilities')

    img_blurred= cv2.medianBlur(img,3)

    ret, img_blurred_thresh = cv2.threshold(img_blurred, 80, 255, cv2.THRESH_BINARY)

    magno_image = cv2.imread(filename+' magno.png')
    if magno_image is None:
        raise FileNotFoundError(filename+' magno.png')
    magno_gray = cv2.cvtColor(magno_image,cv2.COLOR_BGR2GRAY)
    ret, magno_mask = cv2.threshold(magno_gray, 80, 255, cv2.THRESH_BINARY)

    work_img = cv2.bitwise_and(img_blurred_thresh,img_blurred_thresh,mask = magno_mask)
    # Detect blobs my way
    if debug:
        cvShow(work_img, 'Threshold probs blurred magno masked')
    # find local maxima in the probability image by testing which pixels do not change under dilation
    #(dilation replaces each pixel with neighborhood maximum, so only maxima will stay unchanged)
    dilation = cv2.dilate(work_img,cv2.getStructuringElement(cv2.MORPH_RECT,(3,3)),iterations = 1)
    peaks=cv2.compare(work_img,dilation,cv2.CMP_GE)
    if debug:
        cvShow(peaks, 'Peak extraction')
    peaks_masked = cv2.bitwise_and(peaks,peaks,mask = magno_mask)
    if debug:
        cvShow(peaks_masked, 'Peaks magno masked')
    # remove plateau values by erosion (dual of dilation)
    erosion=cv2.erode(work_img,cv2.getStructuringElement(cv2.MORPH_RECT,(3,3)))
    non_plateau=cv2.compare(work_img,erosion,cv2.CMP_GT)
    total=cv2.bitwise_and(non_plateau,peaks_masked)
    if debug:
        cvShow(total,'Plateaus removed')
    # Now determine the centroids of the remaining clusters using scipy
    label_im, nb_labels = ndimage.label(total,cv2.getStructuringElement(cv2.MORPH_RECT,(3,3)))
    # The following removes larger components that are larger than a given size threshold.
    # Note that for the following size claculation the pixel values need to be scaled back to 0/1 from 0/255 (gray values)
    max_size=10
    sizes = ndimage.sum(total/255, label_im, range(nb_labels + 1))
    mask_size = sizes > max_size
    remove_pixel = mask_size[label_im]
    # The line above uses numpy index arrays (see http://docs.scipy.org/doc/numpy/user/basics.indexing.html).
    # Essentially, an array of the same size as the index array (label_im) is generated (copy style) and
    # each element of the result is the value of the element that the corresponding element in the index array
    # indexes in the indexed array.
    # For example, if label_im[3,4]=7, then remove_pixel[3,4]=mask_size[7].
    # The net result is that, since mask_size contains a boolean inidcator for each component number and
    # label_im contains the component numbers, remove_pixel will be a boolean array that contains
    # True for all pixels belonging to clusters that should be removed.
    label_im[remove_pixel] = 0
    label_im, nb_labels = ndimage.label(total,cv2.getStructuringElement(cv2.MORPH_RECT,(3,3)))
    movementCenters=ndimage.measurements.center_of_mass(label_im,label_im, range(1,1+nb_labels))
    #for (y,x) in movementCenters:
    #    cv2.circle(image,(int(x),int(y)),4,(0,255,0),0)

    keypoints = [(x,y) for (y,x) in movementCenters]

    image2=copy.copy(image)

    if debug:
        for (x,y) in keypoints:
            cv2.circle(image,(int(x),int(y)),4,(0,255,0),0)
        cvShow(image,'DIY Keypoints')

        cvShow(work_img, 'review probs for trace_test')

    reducedKeypoints=rejoin_keypoints(work_img, keypoints, debug=debug)

    if debug:
        for (x,y) in reducedKeypoints:
            cv2.circle(image2,(int(x),int(y)),4,(0,255,0),0)
        cvShow(image2,'DIY Keypoints reduced')

    return reducedKeypoints


def run_classifier():
    print "LOADING CLASSIFIER"
    # num_features = 18
    #
    # yes_color=(0,0,255) # red markup for positive class samples
    # no_color=(0,255,0) # green markup for negative class samples (background)
    # black=(0,0,0)

    yes_samples = np.load('yes-samples.npy',None)

    no_samples = np.load('no-samples.npy',None)

    filename = 'classifier_imgs/frame 446'

    # (yes_samples, no_samples)=collect_samples(filename)

    np.random.shuffle(yes_samples)
    np.random.shuffle(no_samples)

    samples=np.concatenate((yes_samples,no_samples))
    class_labels=np.concatenate((np.ones(len(yes_samples)),np.zeros(len(no_samples))))


    # ## Classification using Random Forests
    forest = RandomForestClassifier(n_estimators = 10, max_depth=None, min_samples_split=1, random_state=0)
    forest = forest.fit(samples, class_labels)

    # cvShow(markup_image)
    saveFile = 'Saved Classifier/forest ant'
    joblib.dump(forest, saveFile)

JOIN_DISTANCE = 6
# num_features = 18

#
#   EXEC
#
