from Tracking import *
import copy
import os


""" Drws paths of tracked ants"""

# x coordinate for finishing line:
X_THRESHOLD = 75

joins = []
join_pairs = []

def crossing_check(path):
    """
    receives path, returns True if crossing occurs
    :param path: tuple list (x,y)
    :return: True or none
    """

    def low_check(path):
        """
        checks if any point in path is on the left hand of the threshold line
        :param path: tuple list (x,y)
        :return: boolean value for each point in path
        """
        for point in path:
            yield point[0] < X_THRESHOLD

    def high_check(path):
        """
        checks if any point in path is on the right hand of the threshold line
        :param path: tuple list (x,y)
        :return: boolean value for each point in path
        """
        for point in path:
            yield point[0] >= X_THRESHOLD

    low = low_check(path)
    high = high_check(path)

    if any(low) and any(high):
        return True


def path2polyline(path):
    ptsArray = map((lambda (x,y):[int(round(x)),int(round(y))]), path)
    pts = np.array([ptsArray], np.int32)
    pts = pts.reshape((-1,1,2))
    return pts


def limited_path(specific_ant, start, end):
    """ takes ant object, returns measured path for ant"""
    res = []
    for i in range(len(specific_ant.path)):
        if start <= specific_ant.observation_times[i] <= end:
            res.append(specific_ant.path[i])
    return res


def limited_corrected_path(specific_ant, start, end):
    """ takes ant object, returns kalman corrected path for ant"""
    corrected_path = []
    for i in range(len(specific_ant.path_predicted)):
        if start <= specific_ant.observation_times[i] <= end:
            corrected_path.append(specific_ant.path_predicted[i])

    return corrected_path


def track_exp(exp_nr, time, start, end, path_nos=None, trace_length=1, joined_trace=None, show_joins=False, ):
    counter = 0

    # load ant positions from tracking
    ants = joblib.load('tracking_data/tracked_ants')
    ants = [a for a in ants if len(a.path) > 1]

    joins = []
    curr_cross = []

    print "DRAWING FRAMES"

    for i in range(start, end+1):
        # print status every 50 frames:
        if i % 100 == 0:
            print 'drawing on frame ' + str(i)
        filename = 'img_dump/frame ' + str(i) + ' original.png'
        img = cv2.imread(filename)

        # draw finishing line
        cv2.line(img, (75, 0), (75, 100), (0, 0, 0,), 1)

        past_cross = curr_cross
        curr_cross = []

        if joined_trace:
            cv2.polylines(img,[path2polyline(joined_trace)],False,(255,0,0))

        if path_nos is None:
            for a in ants:
                relevant_path = limited_path(a, i - trace_length, i)
                relevant_corrected_path = limited_corrected_path(a, i - trace_length - 1, i -1)
                cv2.polylines(img, [path2polyline(relevant_path)], False, (0, 0, 255))

                # cv2.polylines(img, [path2polyline(relevant_corrected_path)], False, (0, 255,0))

                # crossing check. further keep track of all current and past crossings, and compare to rule out
                # continuous counting of "dead" tracks
                # also specify use of real track vs kalman corrected one
                if crossing_check(relevant_corrected_path[-2:]):
                    curr_cross.append(relevant_corrected_path[-2:][0])

                    if relevant_corrected_path[-2:][0] not in past_cross:
                        counter += 1
                        cv2.polylines(img, [path2polyline(relevant_path[-2:])], False, (0, 255, 0))

        else:
            for a in [ants[pn] for pn in path_nos]:
                relevant_path = limited_path(a, i-trace_length, i)
                cv2.polylines(img, [path2polyline(relevant_path)],False, (0, 0, 255))


        if show_joins:
            for j in joins:
                cv2.polylines(img,[ path2polyline(j)], False,(255, 255, 0))

        # cv2.imshow('paths',img)
        cv2.putText(img, str(counter), (0, 50), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0))
        img = cv2.resize(img, None, fx=4, fy=4, interpolation=cv2.INTER_CUBIC)

        # only write last few frames
        if LAST_IMAGE_ONLY:
            if i > 719:
                cv2.imwrite('output/img' + str(i) + '.png', img)
        else:
            cv2.imwrite('output/img' + str(i) + '.png', img)
            dir = 'images/' + str(exp_nr) + '_' + str(time) + '_tracked'
            if not os.path.exists(dir):
                os.makedirs(dir)

            cv2.imwrite(dir + '/frame' + str(i) + '.png', img)

        k = cv2.waitKey(1000)
        if k & 0xFF == ord('q'):
            return
        elif k & 0xFF == ord('s'):
            cv2.waitKey(100000)

    print "Counted crossings: ", counter
    return counter


def track_entire_exp(exp_nr, time,  start=3, end=720, path_nos=None, trace_length=1, joined_trace=None, show_joins=False):
    result = track_exp(exp_nr, time, start, end, path_nos=path_nos, trace_length=trace_length, joined_trace=joined_trace, show_joins=show_joins)
    cv2.destroyAllWindows()
    return result


def start_drawing(exp_nr, time, duration):
    result = track_entire_exp(exp_nr, time, start=3, end=duration, trace_length=5)
    return result

# EXEC


LAST_IMAGE_ONLY = False