from Extract_Markup import *
from skimage.feature import structure_tensor, structure_tensor_eigvals
import cv2
from scipy import ndimage

"""Calculating features of ants"""

yes_color =(0, 0, 255)  # red markup for positive class samples
no_color =(0, 255, 0)  # green markup for negative class samples (background)
black = (0, 0, 0)


#
#   FUNCTIONS
#


def coherence(eigenValues):
    """Structure Tensor Coherence"""
    coherence = np.square((eigenValues[0]-eigenValues[1]).astype(np.float32)/(eigenValues[0]+eigenValues[1]))
    return coherence


def structureTensorEigenValues(gray_image, s):
    Axx, Axy, Ayy = structure_tensor(gray_image, sigma=s)
    eigen = structure_tensor_eigvals(Axx, Axy, Ayy)
    return eigen


def compute_all_features(filename):
    # print filename
    global image_orig
    image_orig = cv2.imread(filename+' original.png')
    # Convert BGR to HSV
    global image_hsv
    image_hsv = cv2.cvtColor(image_orig, cv2.COLOR_BGR2HSV)
    # Convert BGR to HSV
    global image_lab
    image_lab = cv2.cvtColor(image_orig, cv2.COLOR_BGR2LAB)
    global image_gauss_blur_03
    image_gauss_blur_03 = ndimage.filters.gaussian_filter(image_orig,0.3)
    global image_gauss_blur_10
    image_gauss_blur_10 = ndimage.filters.gaussian_filter(image_orig,1.0)
    global gray_image
    gray_image = cv2.cvtColor(image_orig, cv2.COLOR_BGR2GRAY)
    global image_laplacian_10
    image_laplacian_10 = ndimage.filters.gaussian_laplace(gray_image,1.0)
    global image_laplacian_35
    image_laplacian_35 = ndimage.filters.gaussian_laplace(gray_image,3.5)
    global image_laplacian_50
    image_laplacian_50 = ndimage.filters.gaussian_laplace(gray_image,5.0)
    global eigen_07
    eigen_07 = structureTensorEigenValues(gray_image, 0.7)
    global eigen_10
    eigen_10 = structureTensorEigenValues(gray_image, 1.0)
    global eigen_35
    eigen_35 = structureTensorEigenValues(gray_image, 3.5)
    global coherence_07
    coherence_07=coherence(eigen_07)
    global coherence_10
    coherence_10=coherence(eigen_10)
    global coherence_35
    coherence_35=coherence(eigen_35)


def feature_vector((x,y), narrow=False):
    features=[]
    global image_gauss_blur_03
    if not narrow:
        features.extend(image_gauss_blur_03[x,y]) # 3 values ?!?
        features.extend(image_gauss_blur_10[x,y]) # 3 values ?!?
        features.append(image_laplacian_10[x,y])
        features.append(image_laplacian_35[x,y])
        features.append(image_laplacian_50[x,y])
        features.append(eigen_07[0][x,y])
        features.append(eigen_07[1][x,y])
        features.append(eigen_10[0][x,y])
        features.append(eigen_10[1][x,y])
        features.append(eigen_35[0][x,y])
        features.append(eigen_35[1][x,y])
        features.append(coherence_07[x,y])
        features.append(coherence_10[x,y])
        features.append(coherence_35[x,y])
    else:
        features.extend(image_hsv[x,y]) # 3 values
    return features


def collect_samples(filename, background=False, narrow=False):
    compute_all_features(filename)
    if background:
        (yes_coords, no_coords) = collect_coordinates(filename+' background.png', yes_color, no_color)
    else:
        (yes_coords, no_coords) = collect_coordinates(filename+' markup.png', yes_color, no_color)
    yes_samples=map(lambda x:feature_vector(x,narrow),yes_coords)
    no_samples=map(lambda x:feature_vector(x,narrow),no_coords)
    return (yes_samples, no_samples)

# Note that all features must be flattened into a single dimensional array (per sample)
def compute_feature_image(filename, narrow=False):
    compute_all_features(filename)
    image_orig = cv2.imread(filename+' original.png')
    if image_orig is None:
        raise FileNotFoundError(image_file_name)
    (h,w) = image_orig.shape[:2]
    feature_vector_length=len(feature_vector((1,1), narrow))
    feature_image=np.empty([h,w,feature_vector_length])
    for x in range(h):
        for y in range(w):
            feature_image[x,y]=feature_vector((x,y), narrow)
    if not narrow:
        np.save(filename+' features',feature_image)
    else:
        np.save(filename + ' narrow features', feature_image)


def run_feature_calc():
    print "CALCULATING FEATURES"
    # num_features = 18
    filename = 'classifier_imgs/frame 446'

    (yes_coords, no_coords) = collect_coordinates(filename + ' markup.png', yes_color, no_color)
    np.save("yes-coords", yes_coords)
    np.save("no-coords", no_coords)
    image_orig = cv2.imread(filename+' original.png')

    image_markup = cv2.imread(filename+' markup.png')

    (yes_samples, no_samples) = collect_samples(filename)
    np.save("yes-samples", yes_samples)
    np.save("no-samples", no_samples)


#
#   EXEC
#