from Classifier_stripped import *
from sklearn.ensemble import RandomForestClassifier
from sklearn.externals import joblib
from datetime import *
import copy
import numpy as np
import cv2
import math

import csv


#
# FUNCTIONS
#
def stale_path(times, now):
    return len(times)>0 and now>times[-1]+2


def decorate(image_A, image_B, old_used, old_unused, new_used, new_unused, new_duplicate):
    old_used=map(to_int_tuple,old_used)
    old_unused=map(to_int_tuple,old_unused)
    new_used=map(to_int_tuple,new_used)
    new_unused=map(to_int_tuple,new_unused)
    new_duplicate=map(to_int_tuple,new_duplicate)
    for x in ant.active_ants:
        if len(x.path)>1:
            cv2.line(image_A,to_int_tuple(x.path[0]),to_int_tuple(x.path[1]),(0,0,255))
            cv2.line(image_B,to_int_tuple(x.path[0]),to_int_tuple(x.path[1]),(0,0,255))
    print "used in green, unused in red"
    print "old as circles, new as squares"
    print "duplicate in aqua - always new"
    for p in old_used: # green circles
        cv2.circle(image_A,p,1,(0,255,0))
        cv2.circle(image_B,p,1,(0,255,0))
    for p in old_unused: # red circles
        cv2.circle(image_A,p,1,(0,0,255))
        cv2.circle(image_B,p,1,(0,0,255))
    for (x, y) in new_used: # green squares
        cv2.rectangle(image_A,(x-1,y-1),(x+1,y+1),(0,255,0),1)
        cv2.rectangle(image_B,(x-1,y-1),(x+1,y+1),(0,255,0),1)
    for (x, y) in new_unused: # red squares
        cv2.rectangle(image_A,(x-1,y-1),(x+1,y+1),(0,0,255),1)
        cv2.rectangle(image_B,(x-1,y-1),(x+1,y+1),(0,0,255),1)
    for (x,y) in new_duplicate: # red squares
        cv2 .rectangle(image_A,(x-1,y-1),(x+1,y+1),(255,255,0),1)
        cv2.rectangle(image_B,(x-1,y-1),(x+1,y+1),(255,255,0),1)


def visualizeTransition(first, second, clf, bg_clf=None):
    filename = 'img_dump/frame '+str(first)
    start_coords = find_ant_coordinates_diy(filename, clf, background_predictor=bg_clf)
    ant.make_ants(start_coords, first)
    old_used, old_unused, new_used, new_unused, new_duplicate = ant.extend_paths(clf, second, bg_classifier=bg_clf)
    image_A = cv2.imread(filename+' original.png')
    if image_A is None:
        raise FileNotFoundError(filename+' original.png')
    filename = 'img_dump/frame '+str(second)+' original.png'
    image_B = cv2.imread(filename)
    if image_B is None:
        raise FileNotFoundError(filename)
    decorate(image_A, image_B, old_used, old_unused, new_used, new_unused, new_duplicate)
    image_A = cv2.resize(image_A, (0,0), fx=3, fy=3)
    image_B = cv2.resize(image_B, (0,0), fx=3, fy=3)
    while True:
        cv2.imshow('Transition', image_A)
        cv2.waitKey(1)
        cv2.imshow('Transition', image_B)
        cv2.waitKey(1)
        if cv2.waitKey(1) & 0xFF == ord('q'):
            cv2.destroyAllWindows()
            break


class ant:
    """the basic ant object"""
    step_lengths = []
    active_ants = []
    closed_ants = []

    def __init__(self, pos, frame_no):
        """
        creates class object for each detected tracked ant. takes position, initialises kalman filter.

        :param pos: x,y position as recorded on first detection.
        :param frame_no: frame of detection
        """
        self.pos = pos
        self.path = [pos]
        self.observation_times = [frame_no]
        self.path_predicted = []
        # self.tmp_measured_pos = np.array((2,1), np.float32)
        # self.pos_predicted = np.array((2,1), np.float32)

        # sets up dimensionality of state and measurement vectors.
        # state vector is 1x4, contains [posX, posY, velX, velY]
        # measurement vector is 1x2, [posX, posY]
        self.kalman = cv2.KalmanFilter(4, 2) # NOTE - Kalman filter is not used in final implementation.
        # measurementMatrix correlates measured values with state values.
        self.kalman.measurementMatrix = np.array([[1, 0, 0, 0], [0, 1, 0, 0]], np.float32)
        # transition matrix equals prediction matrix; shape is [posX, posY, velX, velY]
        self.kalman.transitionMatrix = np.array([[1, 0, 1, 0], [0, 1, 0, 1], [0, 0, 1, 0], [0, 0, 0, 1]], np.float32)

        # assumption about the intensity of noise generated during processing (?)
        # note that in the "Learning OpenCV", example, kalman.processNoiseCov is used instead (multiplied by factor 0.03).
        self.kalman.processNoiseCov = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], np.float32)

        # assumption about the intensity of noise in the measurements
        # self.kalman.measurementNoiseCov = np.array([[1, 0], [0, 1]], np.float32) * 0.03
        if VERBOSE:
            print 'newly initialised ant, initial pos :', self.path


    def get_distance(self, old_pos, new_pos):
        distance = math.hypot(old_pos[0] - new_pos[0], old_pos[1] - new_pos[1])
        return distance


    def move(self, pos, frame_no):
        """
        updates the previous (initial or continuing) track; updates kalman filter, predicts next step
        :param pos:
        :param frame_no:
        :return:
        """
        self.pos = pos
        self.path.append(tuple(pos))
        self.observation_times.append(frame_no)
        self.tmp_measured_pos = np.array([[self.pos[0]], [self.pos[1]]], np.float32)

        if VERBOSE:
            print " -------------- "
            print " newly measured position :", self.tmp_measured_pos


        # kalman filter
        # note we subtract original position from measurements before correcting, and add original position
        # to new prediction
            print 'original pos: ', np.array([[self.path[0][0]], [self.path[0][1]]], np.float32)
            print 'pos for correction: ', self.tmp_measured_pos - np.array([[self.path[0][0]], [self.path[0][1]]], np.float32)

        # correcting position relative to point of first measurement
        self.kalman.correct(self.tmp_measured_pos - np.array([[self.path[0][0]], [self.path[0][1]]], np.float32))

        self.pos_predicted = self.kalman.predict() # + np.array([[self.path[0][0]], [self.path[0][1]]])

        if VERBOSE:
            print 'predicted position: ', str(self.pos_predicted[0, 0]), str(self.pos_predicted[1, 0])
        self.path_predicted.append((self.pos_predicted[0, 0] + self.path[0][0], self.pos_predicted[1, 0] + self.path[0][1]))

        if VERBOSE:
            print 'updated path from measurements: ', self.path
            print 'kalman-corrected path: ', self.path_predicted

        # calculate step length, append to output list step_lengths
        ant.step_lengths.append(self.get_distance(self.path[-2], pos))

    @classmethod
    def make_ants(cls, coords, frame_no):
        """ creates initial set of active_ants """
        ant.active_ants = [ant(pos, frame_no) for pos in coords]
        ant.closed_ants = []

    @classmethod
    def extend_paths(cls, classifier, frame_no, bg_classifier=None):
        """identifies new position, calls on "move" method to append positions to old positions"""
        for this_ant in ant.active_ants:
            if stale_path(this_ant.observation_times, frame_no):
                ant.closed_ants.append(this_ant)
                ant.active_ants.remove(this_ant)

        filename = 'img_dump/frame '+str(frame_no)

        # finds coordinates of ants in image
        new_coords = find_ant_coordinates_diy(filename, classifier, background_predictor=bg_classifier)

        prob_image = cv2.imread(filename+' probabilities.png')
        if prob_image is None:
            raise FileNotFoundError(filename+' probabilities.png')

        magno_image = cv2.imread(filename+' magno.png')
        if magno_image is None:
            raise FileNotFoundError(filename+' magno.png')

        magno_gray = cv2.cvtColor(magno_image, cv2.COLOR_BGR2GRAY)

        max_dist = 7
        old_used = []
        new_used = []
        new_unused = []
        old_unused = []
        new_duplicate = []
        new_coords = list(new_coords)

        # tree for finding nearest neighbours
        kd_tree = spatial.KDTree(new_coords,5)

        old_coords = [x.pos for x in ant.active_ants]
        distances, closest = kd_tree.query(old_coords, distance_upper_bound=max_dist)
        closest = list(closest)
        duplicates = set([x for x in closest if closest.count(x) > 1])
        indices_to_handle = range(len(ant.active_ants))

        while len(indices_to_handle) > 0:
            i = indices_to_handle[0]
            if closest[i] > len(kd_tree.data)-1:
                old_unused.append(old_coords[i])
                indices_to_handle.remove(i)
            else:
                # determining new position
                new_pos = tuple(kd_tree.data[closest[i]])

                # assuming no duplicates exist
                if closest[i] not in duplicates:
                    if path_score(prob_image, to_int_tuple(ant.active_ants[i].pos), to_int_tuple(new_pos), function=np.mean)>80:
                        old_used.append(ant.active_ants[i].pos)
                        new_used.append(new_pos)
                        ant.active_ants[i].move(new_pos, frame_no)
                        new_coords.remove(new_pos)
                        indices_to_handle.remove(i)
                    else:
                        old_unused.append(ant.active_ants[i].pos)
                        new_unused.append(new_pos)
                        new_coords.remove(new_pos)
                        indices_to_handle.remove(i)
                else:
                    # disambiguate duplicate elements here
                    if new_pos in new_coords:  # still available
                        competitors = [j for j in range(len(ant.active_ants)) if closest[j] == closest[i]]
                        scores = [path_score(prob_image, to_int_tuple(ant.active_ants[j].pos), to_int_tuple(new_pos)) for j in competitors]
                        winner_index = scores.index(max(scores))
                        winner = competitors[winner_index]

                        # maybe here we shoudl break the path if there is more than a single deserving competitor
                        if scores[winner_index]>80:  # scores[winner_index]>0
                            old_used.append(ant.active_ants[winner].pos)
                            new_duplicate.append(new_pos)
                            ant.active_ants[winner].move(new_pos, frame_no)
                            new_coords.remove(new_pos)
                            indices_to_handle.remove(winner)
                        else:
                            old_unused.append(ant.active_ants[winner].pos)
                            new_unused.append(new_pos)
                            new_coords.remove(new_pos)
                            indices_to_handle.remove(winner)
                    else:
                        # here we simply put it back into the queue with the next best close point
                        close_distances, close_points = kd_tree.query(ant.active_ants[i].pos,k=20,distance_upper_bound=max_dist)
                        close_points = list(close_points)
                        candidate_indices = [ j for j in close_points if j<len(kd_tree.data) ]
                        usable_indices = [j for j in candidate_indices if tuple(kd_tree.data[j]) in new_coords]
                        if len(usable_indices)>0:
                            closest[i] = usable_indices[0]
                            if closest.count(usable_indices[0]) > 1:
                                # print "assigning duplicate"
                                duplicates.add(usable_indices[0])
                        else:
                            old_unused.append(ant.active_ants[i].pos)
                            indices_to_handle.remove(i)


        # now handle unused new elements
        for new_pos in new_coords:
            new_unused.append(new_pos)
            ant.active_ants.append(ant(new_pos, frame_no))

        return old_used, old_unused, new_used, new_unused, new_duplicate


def show_keypoints_in_context(filename, predictor):
    # this method returns keypoints, not coordinates
    def probs2gray(x):
        low=0
        high=255
        [a,b]=x
        return int(round(a*low+b*high))

    image = cv2.imread(filename+' original.png')
    if image is None:
        raise FileNotFoundError(filename+' original.png')
    compute_feature_image(filename)
    feature_image=np.load(filename+' features.npy',None)
    if feature_image is None:
        raise FileNotFoundError(filename+' features.npy')
    (h,w)=image.shape[:2]
    feature_pixel_list = np.reshape(feature_image,(h*w,num_features))
    pixelprobs=predictor.predict_proba(feature_pixel_list)
    x=np.apply_along_axis(probs2gray,1,pixelprobs)
    img=np.reshape(x.astype('uint8'),(h,w))
    inv = cv2.bitwise_not(img)
    # inv = cv2.medianBlur(inv,3)
    magno_image = cv2.imread(filename+' magno.png')
    if magno_image is None:
        raise FileNotFoundError(filename+' magno.png')
    magno_gray = cv2.cvtColor(magno_image,cv2.COLOR_BGR2GRAY)
    ret, magno_mask = cv2.threshold(magno_gray, 80, 255, cv2.THRESH_BINARY)
    test = cv2.bitwise_and(inv,inv,mask = magno_mask)
    # Detect blobs.
    detector = make_detector()
    keypoints = detector.detect(test)

    im_with_keypoints = cv2.drawKeypoints(test, keypoints, np.array([]), (0,0,255), cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)
    cvShow(im_with_keypoints)


def find_ant_coordinates_diy(filename, predictor, background_predictor=None, debug=False):

    def probs2gray(x):
        low = 0
        high = 255
        [a, b] = x
        return int(round(a * low + b * high))

    image = cv2.imread(filename+' original.png')
    if image is None:
        raise FileNotFoundError(filename+' original.png')

    compute_feature_image(filename)
    feature_image = np.load(filename + ' features.npy', None)

    if feature_image is None:
        raise FileNotFoundError(filename+' features.npy')
    (h,w) = image.shape[:2]
    num_features = len(feature_vector((1,1)))
    feature_pixel_list = np.reshape(feature_image, (h*w, num_features))
    pixelprobs = predictor.predict_proba(feature_pixel_list)
    x = np.apply_along_axis(probs2gray, 1, pixelprobs)
    img = np.reshape(x.astype('uint8'), (h,w))

    if background_predictor:
        compute_feature_image(filename, narrow=True)
        feature_narrow_image=np.load(filename + ' narrow features.npy',None)
        if feature_narrow_image is None:
            raise FileNotFoundError(filename + ' narrow features.npy')
        num_features = len(feature_vector((1, 1), narrow=True))
        feature_narrow_pixel_list = np.reshape(feature_narrow_image, (h*w, num_features))
        pixel_is_background = background_predictor.predict(feature_narrow_pixel_list)
        pixel_is_not_background = np.absolute(np.add(pixel_is_background, -1))
        bg_img=np.reshape(pixel_is_not_background.astype('uint8'), (h, w))

        if debug:
            cvShow(np.multiply(255, bg_img), title="background")

            fg_img = np.reshape(pixel_is_background.astype('uint8'), (h, w))
            modified = np.multiply(img, fg_img)
            cvShow(modified, title="eliminated background probs")

        img = np.multiply(img, bg_img)

    cv2.imwrite(filename+' probabilities.png',img)

    if debug:
        cvShow(img, title='probabilities')

    img_blurred = cv2.medianBlur(img,3)

    ret, img_blurred_thresh = cv2.threshold(img_blurred, 80, 255, cv2.THRESH_BINARY)

    magno_image = cv2.imread(filename +' magno.png')
    if magno_image is None:
        raise FileNotFoundError(filename +' magno.png')
    magno_gray = cv2.cvtColor(magno_image,cv2.COLOR_BGR2GRAY)
    ret, magno_mask = cv2.threshold(magno_gray, 80, 255, cv2.THRESH_BINARY)

    work_img = cv2.bitwise_and(img_blurred_thresh, img_blurred_thresh, mask=magno_mask)
    # Detect blobs my way
    if debug:
        cvShow(work_img, 'Threshold probs blurred magno masked')
    # find local maxima in the probability image by testing which pixels do not change under dilation
    #(dilation replaces each pixel with neighborhood maximum, so only maxima will stay unchanged)
    dilation = cv2.dilate(work_img,cv2.getStructuringElement(cv2.MORPH_RECT,(3,3)),iterations = 1)
    peaks=cv2.compare(work_img,dilation,cv2.CMP_GE)
    if debug:
        cvShow(peaks, 'Peak extraction')
    peaks_masked = cv2.bitwise_and(peaks,peaks,mask = magno_mask)
    if debug:
        cvShow(peaks_masked, 'Peaks magno masked')
    # remove plateau values by erosion (dual of dilation)
    erosion=cv2.erode(work_img,cv2.getStructuringElement(cv2.MORPH_RECT,(3,3)))
    non_plateau=cv2.compare(work_img,erosion,cv2.CMP_GT)
    total=cv2.bitwise_and(non_plateau,peaks_masked)
    if debug:
        cvShow(total,'Plateaus removed')
    # Now determine the centroids of the remaining clusters using scipy
    label_im, nb_labels = ndimage.label(total,cv2.getStructuringElement(cv2.MORPH_RECT,(3,3)))
    # The following removes larger components that are larger than a given size threshold.
    # Note that for the following size claculation the pixel values need to be scaled back to 0/1 from 0/255 (gray values)
    max_size=10
    sizes = ndimage.sum(total/255, label_im, range(nb_labels + 1))
    mask_size = sizes > max_size
    remove_pixel = mask_size[label_im]
    # The line above uses numpy index arrays (see http://docs.scipy.org/doc/numpy/user/basics.indexing.html).
    # Essentially, an array of the same size as the index array (label_im) is generated (copy style) and
    # each element of the result is the value of the element that the corresponding element in the index array
    # indexes in the indexed array.
    # For example, if label_im[3,4]=7, then remove_pixel[3,4]=mask_size[7].
    # The net result is that, since mask_size contains a boolean inidcator for each component number and
    # label_im contains the component numbers, remove_pixel will be a boolean array that contains
    # True for all pixels belonging to clusters that should be removed.
    label_im[remove_pixel] = 0
    label_im, nb_labels = ndimage.label(total,cv2.getStructuringElement(cv2.MORPH_RECT,(3,3)))
    movementCenters = ndimage.measurements.center_of_mass(label_im,label_im, range(1,1+nb_labels))
    #for (y,x) in movementCenters:
    #    cv2.circle(image,(int(x),int(y)),4,(0,255,0),0)

    keypoints = [(x, y) for (y, x) in movementCenters]

    image2=copy.copy(image)

    if debug:
        for (x, y) in keypoints:
            cv2.circle(image, (int(x), int(y)), 4, (0, 255, 0), 0)
        cvShow(image, 'DIY Keypoints')

        cvShow(work_img, 'review probs for trace_test')

    reducedKeypoints = rejoin_keypoints(work_img, keypoints, debug=debug)

    if debug:
        for (x,y) in reducedKeypoints:
            cv2.circle(image2,(int(x),int(y)),4,(0,255,0),0)
        cvShow(image2,'DIY Keypoints reduced')

    return reducedKeypoints


def start_tracking(duration):

    print "TRACKING"

    clf = joblib.load('Saved Classifier/forest ant')
    clf_bg = joblib.load('Saved Classifier/background-forest-446')

    ant_coords_2 = find_ant_coordinates_diy('img_dump/frame 2', clf, background_predictor=clf_bg)

    # initialising ant.active_ants
    ant.make_ants(ant_coords_2, 2)

    #  in each consecutive frame, identify ant positions and extend paths
    for i in range(3, duration):
        if i % 100 == 0:
            print 'processing frame ', str(i)

        ant.extend_paths(clf, i, bg_classifier=clf_bg)

    all_ants = copy.copy(ant.active_ants)
    all_ants.extend(ant.closed_ants)

    #  pickling of kalman objects is not possible; hence replacing them with None. Kalman filter is not needed for
    #  drawing paths; Corrected paths are stored in class attribute corrected_path.
    for a in all_ants:
        a.kalman = None

    joblib.dump(all_ants, 'tracking_data/tracked_ants')

    csv_out = open("step_lengths.csv", 'wb')
    mywriter = csv.writer(csv_out)
    mywriter.writerows(zip(ant.step_lengths))

#
# EXEC
#

VERBOSE = False
