import cv2
import copy
import os

""" extracts frames from target videos."""


# draws the finishing line. use only for manual count images
SAVE_COPY_W_FINISH = True
SCALE_COPY = True



#
#   FUNCTIONS
#


def extract_frames(time, duration, path, crop_coords, exp_nr, rotate_flag):
    """
    takes start
    :param time: start time in seconds
    :param duration:
    :param path:
    :return:
    """

    # opening video file
    cap = cv2.VideoCapture(path)
    cap.set(cv2.CAP_PROP_POS_MSEC, time * 1000)

    # reading frames
    more, frame = cap.read()

    # if rotate_flag is True, rotate image
    if rotate_flag:
        # print frame.shape
        rows, cols, px = frame.shape
        rot_mat = cv2.getRotationMatrix2D((cols / 2, rows / 2), 180, 1)
        frame = cv2.warpAffine(frame, rot_mat, (cols, rows))

    # cropping image
    crop_img = frame[crop_coords[0]:crop_coords[1], crop_coords[2]:crop_coords[3]]

    # allocate a retina instance with input size equal to the one of the loaded image
    retina = cv2.bioinspired.createRetina((crop_img.shape[1], crop_img.shape[0]))

    # retina parameters management methods use sample
    # -> save current (here default) retina parameters to a xml file (you may use it only one time to get the file and modify it)
    retina.write('retinaParams.xml')
    # # -> load retina parameters from a xml file : here we load the default parameters that we just wrote to file
    retina.setup('retinaParams.xml')

    n = 1

    print 'EXTRACTING IMAGES'

    while (cap.isOpened() and more and n <= duration):
        more, frame = cap.read()

        if (more):
            if n % 100 == 0:
                print "extracting frame " + str(n)
            n += 1

            # cropping image
            crop_img = frame[crop_coords[0]:crop_coords[1], crop_coords[2]:crop_coords[3]]

            # run retina on the input image
            retina.run(crop_img)

            # # grab retina outputs
            retinaOut_parvo = retina.getParvo()
            retinaOut_magno = retina.getMagno()

            if cv2.waitKey(1) & 0xFF == ord('q'):
                break

            if n <= duration:
                # normal save operation for current experiment
                cv2.imwrite('img_dump/frame '+str(n)+' magno.png', retinaOut_magno)
                cv2.imwrite('img_dump/frame '+str(n)+' parvo.png', retinaOut_parvo)
                cv2.imwrite('img_dump/frame '+str(n)+' original.png', crop_img)

            if SAVE_COPY_W_FINISH:
                finish_line_img = copy.copy(crop_img)
                cv2.line(finish_line_img, (75, 0), (75, 100), (0, 0, 0,), 1)

                if SCALE_COPY:
                    ##  double size
                    finish_line_img = cv2.resize(finish_line_img, None, fx=4, fy=4, interpolation=cv2.INTER_CUBIC)

                dir = 'images/' + str(exp_nr) + '_' + str(time) +  '_untracked_w_finish'
                if not os.path.exists(dir):
                    os.makedirs(dir)
                cv2.imwrite(dir + '/frame ' + str(n) + '.png', finish_line_img)


    cv2.imwrite('img_dump/frame '+str(n)+' magno.png', retinaOut_magno)
    cv2.imwrite('img_dump/frame '+str(n)+' parvo.png', retinaOut_parvo)
    cv2.imwrite('img_dump/frame '+str(n)+' original.png', crop_img)

    if SAVE_COPY_W_FINISH:
        finish_line_img = copy.copy(crop_img)

        if SCALE_COPY:
            # scales image to 4x size
            finish_line_img = cv2.resize(finish_line_img, None, fx=4, fy=4, interpolation=cv2.INTER_CUBIC)
            cv2.line(finish_line_img, (75, 0), (75, 100), (0, 0, 0,), 1)
            dir = 'images/' + str(exp_nr) + '_' + str(time) + '_untracked_w_finish'
            if not os.path.exists(dir):
                os.makedirs(dir)
            cv2.imwrite(dir + '/frame ' + str(n) + '.png', finish_line_img)
    cap.release()
    cv2.destroyAllWindows()

#
#   EXECUTION
#
